@@ -1,15 +1,16 @@
+from datetime import date as Date, datetime as Datetime, time as Time
 from itertools import count, islice
+from time import localtime
 from warnings import warn
 
 import pg8000
 from pg8000.converters import (
-    BIGINTEGER, BINARY, BOOLEAN, BOOLEAN_ARRAY, BYTES, Binary, CHAR,
-    CHAR_ARRAY, DATE, DATETIME, DECIMAL, DECIMAL_ARRAY, Date, DateFromTicks,
+    BIGINT, BOOLEAN, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY, DATE, DATETIME,
     FLOAT, FLOAT_ARRAY, INET, INT2VECTOR, INTEGER, INTEGER_ARRAY, INTERVAL,
-    JSON, JSONB, MACADDR, NAME, NAME_ARRAY, NULLTYPE, NUMBER, OID, PGInterval,
-    STRING, TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP, TIMESTAMPTZ, Time,
-    TimeFromTicks, Timestamp, TimestampFromTicks, UNKNOWN, UUID_TYPE, VARCHAR,
-    VARCHAR_ARRAY, XID)
+    JSON, JSONB, MACADDR, NAME, NAME_ARRAY, NULLTYPE, NUMERIC, NUMERIC_ARRAY,
+    OID, PGInterval, STRING, TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP,
+    TIMESTAMPTZ, UNKNOWN, UUID_TYPE, VARCHAR, VARCHAR_ARRAY, XID
+)
 from pg8000.core import CoreConnection
 from pg8000.exceptions import DatabaseError, Error, InterfaceError
 
@@ -71,6 +72,88 @@ This property is part of the `DBAPI 2.0 specification
 paramstyle = 'format'
 
 
+BINARY = bytes
+
+
+def PgDate(year, month, day):
+    """Constuct an object holding a date value.
+
+    This function is part of the `DBAPI 2.0 specification
+    <http://www.python.org/dev/peps/pep-0249/>`_.
+
+    :rtype: :class:`datetime.date`
+    """
+    return Date(year, month, day)
+
+
+def PgTime(hour, minute, second):
+    """Construct an object holding a time value.
+
+    This function is part of the `DBAPI 2.0 specification
+    <http://www.python.org/dev/peps/pep-0249/>`_.
+
+    :rtype: :class:`datetime.time`
+    """
+    return Time(hour, minute, second)
+
+
+def Timestamp(year, month, day, hour, minute, second):
+    """Construct an object holding a timestamp value.
+
+    This function is part of the `DBAPI 2.0 specification
+    <http://www.python.org/dev/peps/pep-0249/>`_.
+
+    :rtype: :class:`datetime.datetime`
+    """
+    return Datetime(year, month, day, hour, minute, second)
+
+
+def DateFromTicks(ticks):
+    """Construct an object holding a date value from the given ticks value
+    (number of seconds since the epoch).
+
+    This function is part of the `DBAPI 2.0 specification
+    <http://www.python.org/dev/peps/pep-0249/>`_.
+
+    :rtype: :class:`datetime.date`
+    """
+    return Date(*localtime(ticks)[:3])
+
+
+def TimeFromTicks(ticks):
+    """Construct an objet holding a time value from the given ticks value
+    (number of seconds since the epoch).
+
+    This function is part of the `DBAPI 2.0 specification
+    <http://www.python.org/dev/peps/pep-0249/>`_.
+
+    :rtype: :class:`datetime.time`
+    """
+    return Time(*localtime(ticks)[3:6])
+
+
+def TimestampFromTicks(ticks):
+    """Construct an object holding a timestamp value from the given ticks value
+    (number of seconds since the epoch).
+
+    This function is part of the `DBAPI 2.0 specification
+    <http://www.python.org/dev/peps/pep-0249/>`_.
+
+    :rtype: :class:`datetime.datetime`
+    """
+    return Timestamp(*localtime(ticks)[:6])
+
+
+def Binary(value):
+    """Construct an object holding binary data.
+
+    This function is part of the `DBAPI 2.0 specification
+    <http://www.python.org/dev/peps/pep-0249/>`_.
+
+    """
+    return value
+
+
 def connect(
         user, host='localhost', database=None, port=5432, password=None,
         source_address=None, unix_sock=None, ssl_context=None, timeout=None,
@@ -321,19 +404,6 @@ class Cursor():
                 statement, vals=vals, input_oids=self._input_oids,
                 stream=stream)
 
-            def row_iter(context):
-                it = iter(context.rows)
-                while True:
-                    try:
-                        yield next(it)
-                    except StopIteration as e:
-                        if context is None:
-                            raise ProgrammingError(
-                                "A query hasn't been issued.")
-                        elif len(self._context.columns) == 0:
-                            raise ProgrammingError("no result set")
-                        else:
-                            raise e
             self._row_iter = iter(self._context.rows)
             self._input_oids = None
         except AttributeError as e:
@@ -361,12 +431,34 @@ class Cursor():
             same as the args argument of the :meth:`execute` method.
         """
         rowcounts = []
+        input_oids = self._input_oids
         for parameters in param_sets:
+            self._input_oids = input_oids
             self.execute(operation, parameters)
             rowcounts.append(self._context.row_count)
 
         self._context.row_count = -1 if -1 in rowcounts else sum(rowcounts)
 
+    def callproc(self, procname, parameters=None):
+        args = [] if parameters is None else parameters
+        operation = "CALL " + procname + "(" + \
+            ", ".join(['%s' for _ in args]) + ")"
+
+        try:
+
+            statement, vals = convert_paramstyle('format', operation, args)
+
+            self._context = self._c.execute_unnamed(statement, vals=vals)
+
+            self._row_iter = iter(self._context.rows)
+        except AttributeError as e:
+            if self._c is None:
+                raise InterfaceError("Cursor closed")
+            elif self._c._sock is None:
+                raise InterfaceError("connection is closed")
+            else:
+                raise e
+
     def fetchone(self):
         """Fetch the next row of a query result set.
 
@@ -739,9 +831,9 @@ __all__ = [
     ProgrammingError, Error, OperationalError, IntegrityError, InternalError,
     NotSupportedError, Connection, Cursor, Binary, Date, DateFromTicks, Time,
     TimeFromTicks, Timestamp, TimestampFromTicks, BINARY, PGInterval, STRING,
-    NUMBER, DATETIME, TIME, BOOLEAN, INTEGER, BIGINTEGER, INTERVAL, JSON,
-    JSONB, UNKNOWN, NULLTYPE, ROWID, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY,
-    DATE, DECIMAL, DECIMAL_ARRAY, FLOAT, FLOAT_ARRAY, INET, INT2VECTOR,
-    INTEGER_ARRAY, MACADDR, NAME, NAME_ARRAY, OID, TEXT, TEXT_ARRAY, TIMEDELTA,
-    TIMESTAMP, TIMESTAMPTZ, UUID_TYPE, VARCHAR, VARCHAR_ARRAY, XID
+    DATETIME, TIME, BOOLEAN, INTEGER, BIGINT, INTERVAL, JSON, JSONB, NUMERIC,
+    NUMERIC_ARRAY, NULLTYPE, ROWID, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY,
+    DATE, FLOAT, FLOAT_ARRAY, INET, INT2VECTOR, INTEGER_ARRAY, MACADDR, NAME,
+    NAME_ARRAY, OID, TEXT, TEXT_ARRAY, TIMEDELTA, TIMESTAMP, TIMESTAMPTZ,
+    UNKNOWN, UUID_TYPE, VARCHAR, VARCHAR_ARRAY, XID
 ]
