@@ -1,19 +1,20 @@
+from datetime import date as Date, time as Time
 from itertools import islice
 from warnings import warn
 
 import pg8000
 from pg8000.converters import (
-    BIGINTEGER, BINARY, BOOLEAN, BOOLEAN_ARRAY, BYTES, Binary, CHAR,
-    CHAR_ARRAY, DATE, DATETIME, DECIMAL, DECIMAL_ARRAY, Date, DateFromTicks,
-    FLOAT, FLOAT_ARRAY, INET, INT2VECTOR, INTEGER, INTEGER_ARRAY, INTERVAL,
-    JSON, JSONB, MACADDR, NAME, NAME_ARRAY, NULLTYPE, NUMBER, OID, PGInterval,
-    STRING, TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP, TIMESTAMPTZ, Time,
-    TimeFromTicks, Timestamp, TimestampFromTicks, UNKNOWN, UUID_TYPE, VARCHAR,
-    VARCHAR_ARRAY, XID)
+    BIGINT, BOOLEAN, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY, DATE, FLOAT,
+    FLOAT_ARRAY, INET, INT2VECTOR, INTEGER, INTEGER_ARRAY, INTERVAL, JSON,
+    JSONB, MACADDR, NAME, NAME_ARRAY, NULLTYPE, NUMERIC, NUMERIC_ARRAY, OID,
+    PGInterval, STRING, TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP,
+    TIMESTAMPTZ, UNKNOWN, UUID_TYPE, VARCHAR,
+    VARCHAR_ARRAY, XID, make_params)
 from pg8000.core import CoreConnection
 from pg8000.dbapi import (
-    DataError, IntegrityError, InternalError, NotSupportedError,
-    OperationalError, ProgrammingError, Warning, convert_paramstyle,
+    BINARY, Binary, DataError, DateFromTicks, IntegrityError, InternalError,
+    NotSupportedError, OperationalError, ProgrammingError, TimeFromTicks,
+    Timestamp, TimestampFromTicks, Warning, convert_paramstyle,
 )
 from pg8000.exceptions import DatabaseError, Error, InterfaceError
 
@@ -53,6 +54,10 @@ del get_versions
 __author__ = "Mathieu Fenniak"
 
 
+BIGINTEGER = BIGINT
+DATETIME = TIMESTAMP
+NUMBER = DECIMAL = NUMERIC
+DECIMAL_ARRAY = NUMERIC_ARRAY
 ROWID = OID
 
 
@@ -224,7 +229,9 @@ class Cursor():
             same as the args argument of the :meth:`execute` method.
         """
         rowcounts = []
+        input_oids = self._input_oids
         for parameters in param_sets:
+            self._input_oids = input_oids
             self.execute(operation, parameters)
             rowcounts.append(self._context.row_count)
 
@@ -644,7 +651,7 @@ class PreparedStatement():
 
     def run(self, **vals):
 
-        oids, params = self.con.make_params(self.make_args(vals))
+        oids, params = make_params(self.con.py_types, self.make_args(vals))
 
         try:
             name_bin, row_desc, input_funcs = self.name_map[oids]
