@@ -1,12 +1,12 @@
 from collections import defaultdict
 
 from pg8000.converters import (
-    BIGINTEGER, BINARY, BOOLEAN, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY, DATE,
-    DATETIME, DECIMAL, DECIMAL_ARRAY, FLOAT, FLOAT_ARRAY, INET, INT2VECTOR,
-    INTEGER, INTEGER_ARRAY, INTERVAL, JSON, JSONB, MACADDR, NAME, NAME_ARRAY,
-    NULLTYPE, NUMBER, OID, PGInterval, STRING, TEXT, TEXT_ARRAY, TIME,
-    TIMEDELTA, TIMESTAMP, TIMESTAMPTZ, UNKNOWN, UUID_TYPE, VARCHAR,
-    VARCHAR_ARRAY, XID)
+    BIGINT, BOOLEAN, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY, DATE, FLOAT,
+    FLOAT_ARRAY, INET, INT2VECTOR, INTEGER, INTEGER_ARRAY, INTERVAL, JSON,
+    JSONB, MACADDR, NAME, NAME_ARRAY, NULLTYPE, NUMERIC, NUMERIC_ARRAY, OID,
+    PGInterval, STRING, TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP,
+    TIMESTAMPTZ, UNKNOWN, UUID_TYPE, VARCHAR, VARCHAR_ARRAY, XID, make_params,
+)
 from pg8000.core import CoreConnection
 from pg8000.exceptions import DatabaseError, Error, InterfaceError
 
@@ -184,7 +184,7 @@ class PreparedStatement():
         return self._context.columns
 
     def run(self, stream=None, **params):
-        oids, params = self.con.make_params(self.make_vals(params))
+        oids, params = make_params(self.con.py_types, self.make_vals(params))
 
         try:
             name_bin, columns, input_funcs = self.name_map[oids]
@@ -205,10 +205,10 @@ class PreparedStatement():
 
 
 __all__ = [
-    BINARY, BOOLEAN, BIGINTEGER, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY, DATE,
-    DATETIME, DatabaseError, DECIMAL, DECIMAL_ARRAY, Error, FLOAT, FLOAT_ARRAY,
-    INET, INT2VECTOR, INTEGER, INTEGER_ARRAY, INTERVAL, InterfaceError, JSON,
-    JSONB, MACADDR, NAME, NAME_ARRAY, NULLTYPE, NUMBER, OID, PGInterval,
-    STRING, TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP, TIMESTAMPTZ, UNKNOWN,
+    BYTES, BOOLEAN, BIGINT, BOOLEAN_ARRAY, BYTES, CHAR, CHAR_ARRAY, DATE,
+    DatabaseError, Error, FLOAT, FLOAT_ARRAY, INET, INT2VECTOR, INTEGER,
+    INTEGER_ARRAY, INTERVAL, InterfaceError, JSON, JSONB, MACADDR, NAME,
+    NAME_ARRAY, NULLTYPE, NUMERIC, NUMERIC_ARRAY, OID, PGInterval, STRING,
+    TEXT, TEXT_ARRAY, TIME, TIMEDELTA, TIMESTAMP, TIMESTAMPTZ, UNKNOWN,
     UUID_TYPE, VARCHAR, VARCHAR_ARRAY, XID
 ]
